/*  GCFinder finds conserved gene clusters on multiple genomes based on 
 *    constraining the overall size of a cluster that allows statistical 
 *    significance estimates to be associated with each cluster. 
 * Author: Qingwu Yang
 * Texas A&M University
 * July 6, 2008
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <time.h>

#define EXIT0       fflush(NULL), exit(0)
#define FPFE1(x)    fprintf(stdout, "%s", x), fflush(NULL), exit(0)
#define FPFE2(x, y) fprintf(stdout, x, y), fflush(NULL), exit(0)

//#define GCFINDER_DEBUG

int  IF_CIRCULAR_GENOME = 0; 
int  IF_ORDERED =1;

int  MIN_SET_OUTPUT = 1;      //the minimal size of subsets that will be output
double EVALUE_CUTOFF = 1e-1;  //cutoff for e-values

int MIN_GENOMES=2;           //"minimum number of genomes that a gene cluster must appear"

char tmpf[50]; //= "___tmpclustresults";
char tmpFileName[40]; 

//info()
void info();

//parse commandline
void cmd_parser(int ac, char ** av, int * k, char * fname, char *ofname, char *gfname);

//check if a line is blank
int is_blank(char * s);

//get file names containing input data
void get_file_names(char * fname, char *** gnames, int *n_genomes, int **n_chrs, char ****chrnames);

//do all combinations
void clusterFinder(char **gNames, int nGenomes, int *nChrs, char ***chrNames, int k, char *fName);

//reorganize the output results
void reorganize_output(int ng, char *tfname, char * ofname, char **** geneNames);

//reorganize the output results
void reorganize_output_2(int ng, char *tfname);

//get gene names
void get_gene_names(char * fname, char ***** gnames, int *** gnums, int ** xnums, int ngenomes, char ** genomeNames);

#ifdef GCFINDER_DEBUG
FILE * tim;
#endif

int main(int argc, char *argv[]){
	int numGenomes, *numChrs;
	char **genomeNames, ***chrNames;
	char   inFileName[30], outFileName[30],  gnFileName[30];
	int    distK;

	char **** geneNames;
	int  **   geneNums, *chrNums;
	int  i;

	time_t  timer, time1, time2, time3;

#ifdef GCFINDER_DEBUG
	if( !(tim=fopen("gcfinder.log", "w")) )
		fprintf(stdout, "can't open the time file for writing!\n"), fflush(NULL), exit(0);
	timer=time(NULL);
	fprintf(tim, "Program Starting time: %s\n", asctime(localtime(&timer)));
#endif
	
	//parse commandline
	strcpy(outFileName, "gcluster_result");
	cmd_parser(argc, argv, &distK, inFileName, outFileName, gnFileName);
	strcpy(tmpFileName, outFileName);
	strcat(tmpFileName, "_tmp");
	strcpy(tmpf, "__x__");
	strcat(tmpf, outFileName);

	//get file names containing input data
	get_file_names(inFileName, &genomeNames, &numGenomes, &numChrs, &chrNames);

	//get gene names
	get_gene_names(gnFileName, &geneNames, &geneNums, &chrNums, numGenomes, genomeNames);

	//do all with k=1, 2, ..., distK
	remove(tmpFileName);
	for(i=2; i<distK; i++){
#ifdef GCFINDER_DEBUG
		fprintf(tim, "\n i= %d", i), fflush(NULL);
		time1 = time(NULL);
#endif
		clusterFinder(genomeNames, numGenomes, numChrs, chrNames, i, tmpFileName);
#ifdef GCFINDER_DEBUG
		time2 = time(NULL);
		fprintf(tim, "\nTotal Time: %7.2f mins", (double)difftime(time2, time1)/60), fflush(tim);
#endif
	}

	
	//reorganize the output results
	reorganize_output(numGenomes, tmpFileName, outFileName, geneNames);

	remove(tmpf);
	remove(tmpFileName);

#ifdef GCFINDER_DEBUG
	time3 = time(NULL);
	fprintf(tim, "\nProgram Finish time: %s", asctime(localtime(&time3)));
	fprintf(tim, "\nTotal running time: %7.2f mins", (double)difftime(time3, timer)/60);
	fclose(tim);
#endif

	return 0;
}

typedef struct rnode {
	int   *set;
	double eval;
	int    l;
	int    n;
	char ** s;
	int    *chr;
	int    *low;
	int    *high;
	struct rnode  *next;
} rnode;

//reorganize the output results
void reorganize_output(int ng, char *tfname, char * ofname, char **** geneNames){
	FILE *inf, *outf;
	char  buff[4096];
	int   n, i, j, k, m, x, y, z, t;
	char *p, *p2;

	int  *kinterval, *kprime;

	int dx1, dx2, dx, dx3;

	rnode *head, *tail, *curr, *burr, *durr, *turr, *newnode;

	if( !(kinterval = malloc(ng *sizeof(int))) )
		FPFE1("out of memory for kinterval\n");
	if( !(kprime = malloc(ng * sizeof(int))) )
		FPFE1("out of memory for kprime\n");

	if( !(head = malloc(sizeof(rnode))) )
		FPFE1("out of memory for head\n");
	head->next = NULL;
	n = ng + 2;

	if( !(inf = fopen(tfname, "r")) )
		return;
		//FPFE2("File %s can't be open for reading gcluster\n", tfname);
	if( !(outf = fopen(ofname, "w")) )
		FPFE2("File %s can't be open for writing\n", ofname);

	//read the data
	 while( fgets(buff, 4096, inf) ){
		 if( strstr(buff, "Cluster") ){
			 newnode = malloc(sizeof(rnode));
			 newnode->next = NULL;

			 //read in all data for the new node
			 if( !(newnode->s = malloc(n*sizeof(char *))) )
				 FPFE1("out of memory for newnode->s\n");
			 if( !((newnode->s)[0] = malloc((strlen(buff)+1)*sizeof(char))) )
				 FPFE1("out of memory for newnode->s[0]\n");
			 strcpy((newnode->s)[0], buff);

			 for(i=1; i<n; i++){
				 if( !fgets(buff, 4096, inf) )
					 fprintf(stdout, "data error after the line %s", (newnode->s)[i-1]), fflush(NULL), exit(0);
				 if( strstr(buff, "@@") ){
					 (newnode->s)[i] = NULL;
				 } else {
					if( !((newnode->s)[i] = malloc((strlen(buff)+1)*sizeof(char))) )
						 FPFE1("out of memory for newnode->s[]\n");
					strcpy((newnode->s)[i], buff);
				 }
			 }

			 //determine eval, l, n
			 if( sscanf((newnode->s)[1], " Expect = %le, Size = %d/%*s Appear = %d/%*s", &(newnode->eval), &(newnode->l), &(newnode->n)) !=3){
				 FPFE2("data error in line:\n %s", (newnode->s)[1]);
			 }
			 //determine set
			 if( ( (newnode->set) = malloc(newnode->l * sizeof(int))) ){
				 int xh;//, xi, xj, xp;
				 char *p;

				 //read in the array
				 strcpy(buff, (newnode->s)[0]);
				 if( !(p = strstr(buff, ":")) )
					 FPFE2("data error (no :) in line:\n %s", (newnode->s)[0]);
				 p++;
				 p = strtok(p, " ,");
				 (newnode->set)[0] = atoi(p);
				 xh =1;
				 while( (p = strtok(NULL, " ,")) ){
					(newnode->set)[xh++] = atoi(p);					
				 }
				 if( xh != newnode->l )
					 FPFE2("xh != newnode->l in cluster %s\n", (newnode->s)[0]);

				 /*//sort the array
				 for(xi=0; xi<newnode->l; xi++){
					 xh = (newnode->set)[xi];
					 for(xj=xi+1; xj<newnode->l; xj++){
						 if( (newnode->set)[xj] == xh )
							 FPFE2("(newnode->set)[xj] == xh  in cluster %s\n", (newnode->s)[0]);
						 if( (newnode->set)[xj] < xh ){
							 xh = (newnode->set)[xj];
							 xp = xj;
						 }
					 }
					 if( xh < (newnode->set)[xi] ){
						(newnode->set)[xp] = (newnode->set)[xi];
						(newnode->set)[xi] = xh;
					 }
				 }
				 */
			 } else 
				 FPFE1("out of memory for newnode->set\n");

			 //determine  chr low high
			 if( !( (newnode->chr) = malloc(ng * sizeof(int))) )
				 FPFE1("out of memory for newnode->chr\n");
			 if( !( (newnode->low) = malloc(ng * sizeof(int))) )
				 FPFE1("out of memory for newnode->low\n");
			 if( !( (newnode->high) = malloc(ng * sizeof(int))) )
				 FPFE1("out of memory for newnode->high\n");

			 for(i=2; i<n; i++){
				 if( (newnode->s)[i] == NULL ){
					 (newnode->chr)[i-2] = -1;
					 (newnode->low)[i-2] = -1;
					 (newnode->high)[i-2] = -1;
				 } else {
					 if( sscanf((newnode->s)[i], " %*s (Chr:%d S%d E%d): %*s", &((newnode->chr)[i-2]), &((newnode->low)[i-2]), &((newnode->high)[i-2])) != 3 )
						 FPFE2("data error in line:\n%s", (newnode->s)[i]);
				 }				 
			 }

			 //add the new node to the linked list
			 burr = head;
			 curr = head->next;

			 while( curr != NULL && curr->eval <= newnode->eval ){
				 burr = burr->next;
				 curr = curr->next;
			 }

			 if( curr == NULL ){
				 burr->next = newnode;
				 tail = newnode;
			 } else {
				 burr->next = newnode;
				 newnode->next = curr;
			 }			 
		 }
	 }

//*
	 //remove repeats and subsets
	 if( head->next != NULL && (head->next)->next != NULL ){
		 int yc, yd, y1, y2;
		 burr = head->next;
		 curr = (head->next)->next;
		 while( curr != NULL ){
			 durr = head->next;
			 while( durr != curr ){
				 if( durr->l >= curr->l ){
					 yc = yd = 0;
					 while( yc < curr->l && yd < durr->l ){
						 if( curr->set[yc] == durr->set[yd] ){
							 yc++;
							 yd++;
						 } else if ( curr->set[yc] < durr->set[yd] ){							 
							 break;
						 } else {
							 yd++;
						 }
					 }

					 if( yc == curr->l ){ // possible -- subset or same
						for( y1=0; y1<ng; y1++ ){
							if( curr->chr[y1] != -1 ){
								if( curr->chr[y1] == durr->chr[y1] && curr->low[y1]>=durr->low[y1] && curr->high[y1]<=durr->high[y1] ){
								} else 
									break;
							}
						}
						if( y1 == ng ){ // subset or same
							turr = curr;
							burr->next = curr->next;
							curr = curr->next;

							//free space
							if( turr->set )
								free(turr->set);
							if( turr->chr )
								free(turr->chr);
							if( turr->low )
								free(turr->low);
							if( turr->high )
								free(turr->high);

							for( y2=0; y2<n; y2++ ) 
								if( turr->s[y2] )
									free( turr->s[y2] );
							if( turr->s )
								free(turr->s);
							if( turr )
								free(turr);

							break;
						}

					 }
				 }

				 durr = durr->next;
			 }

			 if ( durr == curr ) {
				 burr=burr->next;
				 curr=curr->next; 
			 }
		 }
	 }
//*/

	 //output new results
	 curr = head->next;
	 while(curr){
		 fprintf(outf, "\n%s", curr->s[0]);
		 fprintf(outf, "%s", curr->s[1]);
		 
		 z=t=0;
		 for(i=0; i<ng; i++){
			 kinterval[i] = kprime[i] = 0;
		 }

		 dx = -1; 

		 for(i=2; i<n; i++){
			 dx3 = -1;

			 if( curr->s[i] ){
				 t++;
				 strcpy(buff, curr->s[i]);  //Cluster_CAL_3 (Chr:3 S77 E82): 233 (77) 234 (78) 235 (82)
				 if( sscanf(buff, " %*s (Chr:%d S%d E%d): %*s ", &j, &x, &y) != 3)
					  FPFE2("data error (no chr no.) in line:\n %s", buff);
				 j--;

				 kinterval[i-2] = y - x;

				 if( !(p = strstr(buff, "):")) )
					 FPFE2("data error (no ):) in line:\n %s", buff);
				 p++;
				 *p='\0';
				 fprintf(outf, "%s: ", buff);
				 p++;

				 dx1 = dx2 = -1;

				 while( (p2 = strstr(p, ")")) ){
					 *p2 = '\0';

					 if( sscanf(p, " %d (%d ", &k, &m) != 2 )
						 FPFE2("data error ( --- ) in line:\n %s", curr->s[i]);

					 fprintf(outf, " %1d (%1d, %1s)", k, m, geneNames[i-2][j][m]);
									 
					 p = p2 + 1;

					 kprime[i-2] ++;

					 //calculate max distance between consecutive genes
					 if ( dx1 < 0 ){
						 dx1 = dx2 = m;
					 } else {
						 dx2 = m;

						 if( dx2 - dx1 > dx3 )
							 dx3 = dx2 - dx1;

						 dx1 = dx2;
					 }
				 }

				 fprintf(outf, "\n");
			 }	

			 if ( dx < dx3 )
				 dx = dx3;

			 //if( curr->s[i] )
			//	 fprintf(outf, "%s", curr->s[i]);
		 }

		 //fprintf(outf, "Max Dist Between Genes: %d\n", dx);

		 /*
		 fprintf(outf, "K: (");
		 for(z=0, i=0; i<ng; i++){
			 if( kinterval[i] >= 0 ){
				 fprintf(outf, " %1d", kinterval[i] +1);

				 if( kinterval[i] > z )
					 z = kinterval[i];
			 }
		 }
		 fprintf(outf, " ) k=%d\n", z  +1);
		 
		 fprintf(outf, "K': (");
		 z=t=0;
		 for(i=0; i<ng; i++){
			 if( kprime[i] > 0 ){
				 fprintf(outf, " %1d", kprime[i]);
				 z += kprime[i];
				 t++;
			 }
		 }
		 fprintf(outf, " ) (average)k'=%d\n", z/t);
		 */

		 curr = curr->next;
	 }
	 
	fclose(inf);
	fclose(outf);

	return;
}

int *tmpx;
char cmdcmd[200];
//do all combinations recurrsively
void recur_comb(int nGenomes, int *nChrs, char ***chrNames, int k, char *fName, int n){
	int i;
	time_t  time1, time2;

	for(i=0; i<nChrs[n]; i++){
		tmpx[n] = i;

		if( n == nGenomes-1 ){
			FILE * outf;
			int j;
			char s[50], t[20];

			s[0] = '\0';

			if( !(outf = fopen(tmpf, "w")) )
				FPFE1("can't open tmpf file for writing!\n");

			for( j=0; j<nGenomes; j++ ){
				fprintf(outf, "%s\n", chrNames[j][tmpx[j]]);

				sprintf(t, "x%1d", tmpx[j]+1);
				strcat(s, t);
			}

			fclose(outf);

			if( IF_ORDERED == 0 ){
				sprintf(cmdcmd, "osynm -i=%s -k=%d -e=%1le -f=%1d -o=%1s -c=%1s -a=%1d", tmpf, k, EVALUE_CUTOFF, IF_CIRCULAR_GENOME, fName, s);
			} else {
				sprintf(cmdcmd, "synm -i=%s -k=%d -e=%1le -f=%1d -o=%1s -c=%1s -a=%1d", tmpf, k, EVALUE_CUTOFF, IF_CIRCULAR_GENOME, fName, s, MIN_GENOMES);
			}
			//else if( IF_ORDERED == 1 ) {
			//	sprintf(cmdcmd, "syn -i=%s -k=%d -e=%1le -f=%1d -o=%1s -c=%1s ", tmpf, k, EVALUE_CUTOFF, IF_CIRCULAR_GENOME, fName, s);
			//} else if( IF_ORDERED == 2 ) {
			//	sprintf(cmdcmd, "synm -i=%s -k=%d -e=%1le -f=%1d -o=%1s -c=%1s ", tmpf, k, EVALUE_CUTOFF, IF_CIRCULAR_GENOME, fName, s);
			//}

#ifdef GCFINDER_DEBUG
			fprintf(tim, "\n%s", cmdcmd), fflush(tim);
			time1 = time(NULL);
#endif

			system(cmdcmd);

#ifdef GCFINDER_DEBUG
			time2 = time(NULL);
			fprintf(tim, "\n\t %7.2f mins", (double)difftime(time2, time1)/60), fflush(tim);			
#endif
		} else {
			recur_comb(nGenomes, nChrs, chrNames, k, fName, n+1);
		}

		//remove repeats and subsets
		//if( n == nGenomes-3 )
		//	reorganize_output_2(nGenomes, tmpFileName);
	}
}

//do all combinations
void clusterFinder(char **gNames, int nGenomes, int *nChrs, char ***chrNames, int k, char *fName){
	if( !(tmpx = malloc(nGenomes * sizeof(int))) )
		FPFE1("out of memory for tmpx\n");
	//remove(fName);
	recur_comb(nGenomes, nChrs, chrNames, k, fName, 0);

	return;
}

//convert '\n' or 'EOF' into '\0'
void remove_NEOF(char * s){
	while( *s != '\0' && *s != '\n' && *s != EOF )  s++;
	if( *s != '\0' )  *s = '\0';
}

//get file names containing input data
void get_file_names(char * fname, char *** gnames, int *n_genomes, int **n_chrs, char ****chrnames){
	FILE *inf;
	char buff[4096], *ptr;
	char ** gns, ***xnames;
	int i, n, m, *nxs;

	if( !(inf = fopen(fname, "r")) )
		FPFE2("File %1s can't be opened for reading", fname);

	//get the number of genomes
	n=0;
	while( fgets(buff, 4096, inf) ){
		if( strstr(buff, ">") )
			n++;
	}

	*n_genomes = n;
	if( !(gns = malloc(n * sizeof(char *))) )
		FPFE1("out of memory for *genomenames!\n");
	if( !(nxs = malloc(n * sizeof(int))) )
		FPFE1("out of memory for nxs!\n");
	for(i=0; i<n; i++)
		nxs[i]=0;

	if( !(xnames = malloc(n* sizeof(char **))) )
		FPFE1("out of memory for xnames**!\n");

	//read in gnames and count numbers of chrs
	rewind(inf);
	n=0;
	while( fgets(buff, 4096, inf) ){
		if( !is_blank(buff) ){
			if( (ptr=strstr(buff, ">")) ){
				remove_NEOF(ptr);
				ptr++;
				if( !( gns[n] = malloc( (strlen(ptr)+1) * sizeof(char))) )
					FPFE1("out of memory for gns[]!\n");
				strcpy(gns[n], ptr);
				n++;
			} else {
				if( n == 0 )
					FPFE2("wrong format in input file: %s\n", fname);
				nxs[n-1] ++;
			}
		}
	}

	for(i=0; i<n; i++)
		if( !(xnames[i] = malloc(nxs[i] * sizeof(char *))) )
			FPFE1("out of memory for xnames*!\n");

	//read in chrnames
	rewind(inf);
	n=0;
	while( fgets(buff, 4096, inf) ){
		if( !is_blank(buff) ){
			if( strstr(buff, ">") ){
				m=0;
				n++;
			} else {
				ptr = buff;
				while( !isalnum(*ptr) ) ptr++;
				remove_NEOF(ptr);
				if( !( xnames[n-1][m] = malloc( (strlen(ptr)+1) * sizeof(char))) )
					FPFE1("out of memory for xnames[][]!\n");
				strcpy(xnames[n-1][m], ptr);
				m++;
			}
		}
	}

	*n_chrs   = nxs;
	*gnames = gns;
	*chrnames    = xnames;
	fclose(inf);
	return;
}

//check if a line is blank
int is_blank(char * s){
	while( *s != '\0' && *s != EOF ){
		if( isalnum(*s) )
			return 0;
		s++;
	}
	return 1;
}

//info()
void info(){
	fprintf(stdout, "\ngcfinder ");
	fprintf(stdout, "\n\t -h= \"file name specifying files that define the homologous groups\" ");  
	fprintf(stdout, "\n\t -g= \"file name specifying files that give the gene names\" ");
	
	fprintf(stdout, "\n\t -t= \"type of gene clusters\" ");
	fprintf(stdout, "\n\t\t 0 -- ordered clusters ");
	fprintf(stdout, "\n\t\t 1 -- unordered clusters ");
	//fprintf(stdout, "\n\t\t 1 -- unordered clusters, each gene belongs to one homologous group ");
	//fprintf(stdout, "\n\t\t 2 -- unordered clusters, each gene may belong to many homologous groups ");

	fprintf(stdout, "\n\t -u= \"type of genomes\" "); 
	fprintf(stdout, "\n\t\t 0 -- linear ");
	fprintf(stdout, "\n\t\t 1 -- circular ");

	fprintf(stdout, "\n\t -a= \"minimum number of genomes that a gene cluster must appear\" ");
	fprintf(stdout, "\n\t -d= \"maximum size of gene clusters\" ");
	fprintf(stdout, "\n\t -e= \"e-value cutoff, only gene clusters with lower e-value are returned\" ");
	
	fprintf(stdout, "\n\t -o= \"output file name\"\n\n");
	
	fflush(NULL);
	exit(0);
}

//parse commandline
void cmd_parser(int ac, char ** av, int * k, char * fname, char *ofname, char *gfname){
	if( ac < 3 ){
		info();
	} else { 
		ac--;
		while( ac > 0 ){
			if( strstr(av[ac], "-h") ){
				if( sscanf(av[ac], "-h=%s", fname) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			} else if( strstr(av[ac], "-a") ){
				if( sscanf(av[ac], "-a=%d", &MIN_GENOMES) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			} else if( strstr(av[ac], "-u=") ){
				if( sscanf(av[ac], "-u=%d", &IF_CIRCULAR_GENOME) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			}  else if( strstr(av[ac], "-t=") ){
				if( sscanf(av[ac], "-t=%d", &IF_ORDERED) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			} else if( strstr(av[ac], "-g=") ){
				if( sscanf(av[ac], "-g=%s", gfname) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			} else if( strstr(av[ac], "-o=") ){
				if( sscanf(av[ac], "-o=%s", ofname) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			} else if( strstr(av[ac], "-d=") ){
				if( sscanf(av[ac], "-d=%d", k) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			} else if( strstr(av[ac], "-n=") ){
				if( sscanf(av[ac], "-n=%d", &MIN_SET_OUTPUT) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			} else if( strstr(av[ac], "-e=") ){
				if( sscanf(av[ac], "-e=%lf", &EVALUE_CUTOFF) != 1 ){
					fprintf(stdout, "wrong command line input: %s\n", av[ac]);
					info();
				}
			} else {
				fprintf(stdout, "wrong command line input: %s\n", av[ac]);
				info();
			}
			ac--;
		}
	}

	return;
}

//get gene names
void get_gene_names(char * fname, char ***** gnames, int *** gnums, int ** xnums, int ngenomes, char ** genomeNames){
	FILE *inf, * inf2;
	char buff[4096], *ptr;
	char **** gs;
	int  **   gn;
	int i, n, m, *nxs, max;
	char str[100];

	if( !(inf = fopen(fname, "r")) )
		FPFE2("File %1s can't be opened for reading", fname);

	//get the number of genomes
	n=0;
	while( fgets(buff, 4096, inf) ){
		if( strstr(buff, ">") )
			n++;
	}

	if( n != ngenomes )
		FPFE1("n != ngenomes\n");

	if( !(gs = malloc(n * sizeof(char ***))) )
		FPFE1("out of memory for gs***!\n");
	if( !(gn = malloc(n * sizeof(int *))) )
		FPFE1("out of memory for gn* !\n");
	if( !(nxs = malloc(n * sizeof(int))) )
		FPFE1("out of memory for nxs !\n");

	for(i=0; i<n; i++)
		nxs[i] = 0;

	//test genome names and count numbers of chrs
	rewind(inf);
	n=0;
	while( fgets(buff, 4096, inf) ){
		if( !is_blank(buff) ){
			if( (ptr=strstr(buff, ">")) ){
				remove_NEOF(ptr);
				ptr++;
				strcpy(str, ptr);
				if( strcmp(str, genomeNames[n]) ) 
					FPFE1("gene files and cluster files are in different order?\n");
				n++;
			} else {
				if( n == 0 )
					FPFE2("wrong format in input file: %s\n", fname);
				nxs[n-1] ++;
			}
		}
	}

	for(i=0; i<ngenomes; i++)
		if( !(gs[i] = malloc(nxs[i] * sizeof(char **))) )
			FPFE1("out of memory for gs**!\n");
	for(i=0; i<ngenomes; i++)
		if( !(gn[i] = malloc(nxs[i] * sizeof(int))) )
			FPFE1("out of memory for gn !\n");

	//read in gene names
	rewind(inf);
	n=-1;
	while( fgets(buff, 4096, inf) ){
		if( !is_blank(buff) ){
			if( strstr(buff, ">") ){
				m=0;
				n++;
			} else {
				ptr = buff;
				while( !isalnum(*ptr) ) ptr++;
				remove_NEOF(ptr);
				strcpy(str, ptr);

				if( !(inf2 = fopen( str, "r" )) )
					FPFE2("File %s can't be opened for reading!\n", str);
				
				gn[n][m]=0;
				max = 0;
				while( fgets(buff, 4096, inf2) ){
					if( !is_blank(buff) ){
						if( sscanf(buff, " %d %*s ", &i) != 1 )
							FPFE2("data error in gFile %s\n", str);
						if( i > max )
							max = i;
					}
				}
				gn[n][m] = max;

				if( !(gs[n][m] = malloc( (max+1) * sizeof(char*))) )
					FPFE1("out of memory for gs*!\n");

				for(i=0; i<=max; i++)
					gs[n][m][i] = NULL;

				rewind(inf2);
				while( fgets(buff, 4096, inf2) ){
					if( !is_blank(buff) ){
						if( sscanf(buff, " %d %s ", &i, str) != 2 )
							FPFE2("data error in gFile %s\n", str);

						if( !(gs[n][m][i] = malloc( (strlen(str)+1) * sizeof(char))) )
							FPFE1("out of memory for gs!\n");
						strcpy(gs[n][m][i], str);						
					}
				}

				fclose(inf2);

				m++;
			}
		}
	}

	fclose(inf);

	*gnames = gs; 
	*gnums = gn; 
	*xnums = nxs;
	return;
}


//reorganize the output results
void reorganize_output_2(int ng, char *tfname){
	FILE *inf, *outf;
	char  buff[4096];
	int   n, i, y2;

	rnode *head, *tail, *curr, *burr, *durr, *turr, *newnode;

	if( !(head = malloc(sizeof(rnode))) )
		FPFE1("out of memory for head\n");
	head->next = NULL;
	n = ng + 2;

	if( !(inf = fopen(tfname, "r")) )
		return;

	//read the data
	 while( fgets(buff, 4096, inf) ){
		 if( strstr(buff, "Cluster") ){
			 newnode = malloc(sizeof(rnode));
			 newnode->next = NULL;

			 //read in all data for the new node
			 if( !(newnode->s = malloc(n*sizeof(char *))) )
				 FPFE1("out of memory for newnode->s\n");
			 if( !((newnode->s)[0] = malloc((strlen(buff)+1)*sizeof(char))) )
				 FPFE1("out of memory for newnode->s[0]\n");

			 strcpy((newnode->s)[0], buff);

			 for(i=1; i<n; i++){
				 if( !fgets(buff, 4096, inf) )
					 fprintf(stdout, "data error after the line %s", (newnode->s)[i-1]), fflush(NULL), exit(0);
					 //FPFE2("data error after the line %s", (newnode->s)[i-1]);
				 if( strstr(buff, "@@") ){
					 (newnode->s)[i] = NULL;
				 } else {
					if( !((newnode->s)[i] = malloc((strlen(buff)+1)*sizeof(char))) )
						 FPFE1("out of memory for newnode->s[]\n");
					strcpy((newnode->s)[i], buff);
				 }
			 }

			 //determine eval, l, n
			 if( sscanf((newnode->s)[1], " Expect = %le, Size = %d/%*s Appear = %d/%*s", &(newnode->eval), &(newnode->l), &(newnode->n)) !=3){
				 FPFE2("data error in line:\n %s", (newnode->s)[1]);
			 }
			 //determine set
			 if( ( (newnode->set) = malloc(newnode->l * sizeof(int))) ){
				 int xh;//, xi, xj, xp;
				 char *p;

				 //read in the array
				 strcpy(buff, (newnode->s)[0]);
				 if( !(p = strstr(buff, ":")) )
					 FPFE2("data error (no :) in line:\n %s", buff);
				 p++;
				 p = strtok(p, " ,");
				 (newnode->set)[0] = atoi(p);
				 xh =1;
				 while( (p = strtok(NULL, " ,")) ){
					(newnode->set)[xh++] = atoi(p);					
				 }
				 if( xh != newnode->l )
					 FPFE2("xh != newnode->l in cluster %s\n", (newnode->s)[0]);
			 } else 
				 FPFE1("out of memory for newnode->set\n");

			 //determine  chr low high
			 if( !( (newnode->chr) = malloc(ng * sizeof(int))) )
				 FPFE1("out of memory for newnode->chr\n");
			 if( !( (newnode->low) = malloc(ng * sizeof(int))) )
				 FPFE1("out of memory for newnode->low\n");
			 if( !( (newnode->high) = malloc(ng * sizeof(int))) )
				 FPFE1("out of memory for newnode->high\n");

			 for(i=2; i<n; i++){
				 if( (newnode->s)[i] == NULL ){
					 (newnode->chr)[i-2] = -1;
					 (newnode->low)[i-2] = -1;
					 (newnode->high)[i-2] = -1;
				 } else {
					 if( sscanf((newnode->s)[i], " %*s (Chr:%d S%d E%d): %*s", &((newnode->chr)[i-2]), &((newnode->low)[i-2]), &((newnode->high)[i-2])) != 3 )
						 FPFE2("data error in line:\n%s", (newnode->s)[i]);
				 }				 
			 }

			 //add the new node to the linked list
			 burr = head;
			 curr = head->next;

			 while( curr != NULL && curr->eval <= newnode->eval ){
				 burr = burr->next;
				 curr = curr->next;
			 }

			 if( curr == NULL ){
				 burr->next = newnode;
				 tail = newnode;
			 } else {
				 burr->next = newnode;
				 newnode->next = curr;
			 }			 
		 }
	 }

	 //remove repeats and subsets
	 if( head->next != NULL && (head->next)->next != NULL ){
		 int yc, yd, y1;
		 burr = head->next;
		 curr = (head->next)->next;
		 while( curr != NULL ){
			 durr = head->next;
			 while( durr != curr ){
				 if( durr->l >= curr->l ){
					 yc = yd = 0;
					 while( yc < curr->l && yd < durr->l ){
						 if( curr->set[yc] == durr->set[yd] ){
							 yc++;
							 yd++;
						 } else if ( curr->set[yc] < durr->set[yd] ){							 
							 break;
						 } else {
							 yd++;
						 }
					 }

					 if( yc == curr->l ){ // possible -- subset or same
						for( y1=0; y1<ng; y1++ ){
							if( curr->chr[y1] != -1 ){
								if( curr->chr[y1] == durr->chr[y1] && curr->low[y1]>=durr->low[y1] && curr->high[y1]<=durr->high[y1] ){
								} else 
									break;
							}
						}
						if( y1 == ng ){ // subset or same
							turr = curr;
							burr->next = curr->next;
							curr = curr->next;

							//free space
							if( turr->set )
								free(turr->set);
							if( turr->chr )
								free(turr->chr);
							if( turr->low )
								free(turr->low);
							if( turr->high )
								free(turr->high);

							for( y2=0; y2<n; y2++ ) 
								if( turr->s[y2] )
									free( turr->s[y2] );
							if( turr->s )
								free(turr->s);
							if( turr )
								free(turr);

							break;
						}

					 }
				 }

				 durr = durr->next;
			 }

			 if ( durr == curr ) {
				 burr=burr->next;
				 curr=curr->next; 
			 }
		 }
	 }

	 //remove old version
	 fclose(inf);
	 remove(tfname);

	 if( !(outf = fopen(tfname, "w")) )
		return;

	 //output new results
	 curr = head->next;
	 while(curr){
		 turr = curr;
		 for(i=0; i<n; i++)
			if( curr->s[i] ){
				fprintf(outf, "%s", curr->s[i]);
			} else {
				fprintf(outf, "@@\n");
			}
		 fprintf(outf, "\n\n");
		 curr = curr->next;

		 //free space
		if( turr->set )
			free(turr->set);
		if( turr->chr )
			free(turr->chr);
		if( turr->low )
			free(turr->low);
		if( turr->high )
			free(turr->high);

		for( y2=0; y2<n; y2++ ) 
			if( turr->s[y2] )
				free( turr->s[y2] );
		if( turr->s )
			free(turr->s);
		if( turr )
			free(turr);
	 }
	 
	fclose(outf);

	return;
}
